// Example pys60 audio extension for CCRMA Mobile Music Workshop, 2007-11-01
// Copyright (c) Nokia, written by Jarno Seppänen

/* Include Python.h but make sure _DEBUG is not on. */
#ifdef _DEBUG
# undef _DEBUG
# include <Python.h>
# define _DEBUG
#else
# include <Python.h>
#endif

#include <math.h> // fabs
#include <stdlib.h> // malloc, free
#include <symbian_python_ext_util.h> // SPyErr_SetFromSymbianOSErr, SPyAddGlobalString
#include <e32std.h> // KErrNone, TDllReason
#include <mda/common/audio.h> // MMdaAudioOutputStreamCallback
#include <mdaaudiooutputstream.h> // CMdaAudioOutputStream
#include <RDAccelerometer.h>
#include <RDAccelerometerObserver.h>

static const int BUFLEN=4096/2/2;  // samples in a 4K stereo buffer

#include "wavetable.cpp" // WAVETABLE, WAVETABLE_LEN

class Shake : public CBase,
              public virtual MMdaAudioOutputStreamCallback,
              public virtual MRDAccelerometerObserver
{
public:
    Shake();
    virtual ~Shake();

    // MMdaAudioOutputStreamCallback
    virtual void MaoscPlayComplete(TInt error);
    virtual void MaoscBufferCopied(TInt error, const TDesC8& buffer);
    virtual void MaoscOpenComplete(TInt error);

    // MRDAccelerometerObserver
    virtual void HandleAccelerationL(TInt aX, TInt aY, TInt aZ);

    // Python interface methods
    virtual int Open();
    virtual void Close();
    virtual void SetVolume(double volume);
    virtual void SetFrequency(int axis, double frequency);
    virtual void SetPitchRange(int axis, double octaves);
    virtual void SetGain(int axis, double gain);

protected:
    void FillBuffer(int error);

private:
    CMdaAudioOutputStream* iStream;
    int iAudioError;
    short* iBuffer;
    TPtrC8 iBufferPtr;
    double iPitch[3];      // zero-G base pitch (Hertz)
    double iPitchRange[3]; // one-G pitch shift (octaves)
    int iGain[3];       // oscillator gains
    int iStep[3];       // filtered wavetable step (15.16)
    int iTargetStep[3]; // target (unfiltered) wavetable step (15.16)
    int iPos[3];        // wavetable position (15.16)

    CRDAccelerometer* iAccelerometer;
    TReal iX;
    TReal iY;
    TReal iX1;
    TReal iY1;
    TReal iZ0;
    TReal iZ1;
    TReal iZ2;
};

Shake::Shake()
    : iStream(0),
      iAudioError(0),
      iBuffer(0),
      iAccelerometer(0),
      iX(0),
      iY(0),
      iX1(0),
      iY1(0),
      iZ0(0),
      iZ1(0),
      iZ2(0)
{}

Shake::~Shake()
{
    Close();
}

void 
Shake::MaoscBufferCopied(TInt error, const TDesC8& /*buffer*/)
{
    FillBuffer(error);
}

void 
Shake::MaoscOpenComplete(TInt error)
{
    FillBuffer(error);
}

void 
Shake::MaoscPlayComplete(TInt error)
{
    FillBuffer(error);
}

void
Shake::HandleAccelerationL(TInt aX, TInt aY, TInt aZ)
{
    const TReal Kscale = 1.0/64;
    //const TReal KPole1 = 0.97;//pole of lowpass filter applied to X and Y
    //const TReal KPole2 = 0.90;//pole of lowpass filter applied to Z
    //const TReal KPole3 = 0.75;//pole of highpass filter applied to Z
    const TReal KPole1 = 0.50;
    const TReal KPole2 = 0.40;
    const TReal KPole3 = 0.20;

    //single-pole lowpass-filtering of X and Y
    iX = KPole1 * iX1 + ( 1.0 - KPole1 ) * ( Kscale * aX );
    iY = KPole1 * iY1 + ( 1.0 - KPole1 ) * ( Kscale * aY );

    //single-pole lowpass-filtering of Z
    iZ0 = KPole2 * iZ1 + ( 1.0 - KPole2 ) * ( Kscale * aZ );
    //1st order highpass-filtering of Z (zero at 1)
    iZ2 = KPole3 * iZ2 + ( 1.0 - KPole3 ) * ( iZ0 - iZ1 );

    //update variables for use at next sample
    iX1 = iX;
    iY1 = iY;
    iZ1 = iZ0;

    // update sound frequency
    static const double s=1.0/48000.0*WAVETABLE_LEN*65536;
    iTargetStep[0]=int(iPitch[0]*s*pow(2, iPitchRange[0]*iX));
    iTargetStep[1]=int(iPitch[1]*s*pow(2, iPitchRange[1]*iY));
    iTargetStep[2]=int(iPitch[2]*s*pow(2, iPitchRange[2]*iZ2));
}

int
Shake::Open()
{
    // initialize oscillators
    iPitch[0]=440.0; // Hertz
    iPitch[1]=440.0;
    iPitch[2]=440.0;
    iPitchRange[0]=1.0; // octaves
    iPitchRange[1]=1.0;
    iPitchRange[2]=1.0;
    iGain[0]=int(0.3*65536);
    iGain[1]=int(0.3*65536);
    iGain[2]=int(0.3*65536);
    iStep[0]=int(iPitch[0]/48000.0*WAVETABLE_LEN*65536);
    iStep[1]=int(iPitch[1]/48000.0*WAVETABLE_LEN*65536);
    iStep[2]=int(iPitch[2]/48000.0*WAVETABLE_LEN*65536);
    iTargetStep[0]=iStep[0];
    iTargetStep[1]=iStep[1];
    iTargetStep[2]=iStep[2];
    iPos[0]=0;
    iPos[1]=0;
    iPos[2]=0;

    // open Symbian audio stream
    TMdaAudioDataSettings settings;
    settings.iChannels=TMdaAudioDataSettings::EChannelsStereo;
    settings.iSampleRate=TMdaAudioDataSettings::ESampleRate48000Hz;
    settings.iVolume=settings.iMaxVolume>>1;

    // NB. do not exceed 4096 bytes here
    iBuffer=(TInt16*)malloc(BUFLEN*2*sizeof(TInt16));
    if (!iBuffer) {
        return KErrNoMemory;
    }
    iBufferPtr.Set(TPtrC8((TUint8*)(iBuffer), BUFLEN*2*sizeof(TInt16)));

    // open audio output, start the FillBuffer callbacks
    int error;
    TRAP(error, iStream=CMdaAudioOutputStream::NewL(*this));
    if (error) {
        free(iBuffer);
        iBuffer=0;
        return error;
    }
    iStream->Open(&settings);

    // open accelerometer, start the HandleAcceleration callbacks
    TRAP(error, iAccelerometer=CRDAccelerometer::NewL(*this));
    if (error) {
        iStream->Stop();
        delete iStream;
        iStream=0;
        free(iBuffer);
        iBuffer=0;
        return error;
    }

    return KErrNone;
}

void
Shake::Close()
{
    delete iAccelerometer;
    iAccelerometer=0;
    iStream->Stop();
    delete iStream;
    iStream=0;
    free(iBuffer);
    iBuffer=0;
}

void
Shake::SetVolume(double vol)
{
    double maxvol=iStream->MaxVolume();
    vol=vol*maxvol+0.5;
    vol=0.5*(vol+maxvol-::fabs(vol-maxvol)); // min(vol,maxvol)
    iStream->SetVolume(int(vol));
}

void
Shake::SetFrequency(int axis, double freq)
{
    if (axis<0 || axis>2) {
        return;
    }
    iPitch[axis]=freq;
}

void
Shake::SetPitchRange(int axis, double octaves)
{
    if (axis<0 || axis>2) {
        return;
    }
    iPitchRange[axis]=octaves;
}

void
Shake::SetGain(int axis, double gain)
{
    if (axis<0 || axis>2) {
        return;
    }
    iGain[axis]=int(gain*65536);
}

void
Shake::FillBuffer(int error)
{
    if (error) {
        iAudioError=error;
        return;
    }
    for (int i=0; i<BUFLEN; i++) {
        int x=0;
        // mix three sines
        x=(iGain[0]*WAVETABLE[iPos[0]>>16])>>16;  // 16x16->16
        x+=(iGain[1]*WAVETABLE[iPos[1]>>16])>>16; // 16x16->16
        x+=(iGain[2]*WAVETABLE[iPos[2]>>16])>>16; // 16x16->16
        iBuffer[2*i]=x;   // left channel
        iBuffer[2*i+1]=x; // right channel
        // update wavetable playback positions
        //assert(WAVETABLE_LEN<32768);
        const int per=WAVETABLE_LEN*65536;
        iPos[0]=(iPos[0]+iStep[0])%per;
        iPos[1]=(iPos[1]+iStep[1])%per;
        iPos[2]=(iPos[2]+iStep[2])%per;
        // update wavetable playback steps smoothly
        const int c0=0.999423*65536; // 25 ms half-life (1200 samples@48 kHz)
        // to change half-life, recompute coefficient from c=0.5^(1/tau/48000)
        // where tau is half-life in seconds
        const int c1=65536-c0;
        //assert(c1>0);
        iStep[0]=(c0*TInt64(iStep[0])+c1*TInt64(iTargetStep[0]))>>16; // 16x32->32
        iStep[1]=(c0*TInt64(iStep[1])+c1*TInt64(iTargetStep[1]))>>16; // 16x32->32
        iStep[2]=(c0*TInt64(iStep[2])+c1*TInt64(iTargetStep[2]))>>16; // 16x32->32
    }
    iStream->WriteL(iBufferPtr);
}

extern "C" {

struct Shake_PyObject {
    PyObject_VAR_HEAD
    Shake* shake;
};

static PyObject* Shake_PyObject_new(PyTypeObject* type, PyObject* args, PyObject* kwds);
static void Shake_PyObject_dealloc(Shake_PyObject*);
static PyObject* Shake_Open(Shake_PyObject*, PyObject*);
static PyObject* Shake_Close(Shake_PyObject*);
static PyObject* Shake_SetVolume(Shake_PyObject*, PyObject*);
static PyObject* Shake_SetFrequency(Shake_PyObject*, PyObject*);
static PyObject* Shake_SetPitchRange(Shake_PyObject*, PyObject*);
static PyObject* Shake_SetGain(Shake_PyObject*, PyObject*);

static const PyMethodDef shake_methods[] = {
    {NULL, NULL}
};

static const PyMethodDef Shake_methods[] = {

    {"Open", (PyCFunction)Shake_Open, METH_VARARGS, "Open()\n"},
    {"Close", (PyCFunction)Shake_Close, METH_NOARGS, "Close()\n"},
    {"SetVolume", (PyCFunction)Shake_SetVolume, METH_VARARGS, "SetVolume(vol: float)\nVolume between 0 (silent) and 1 (maximum)."},
    {"SetFrequency", (PyCFunction)Shake_SetFrequency, METH_VARARGS, "SetFrequency(axis: int, freq: float)\nAxis X=0, Y=1, Z=2, frequency in Hertz (for zero G)"},
    {"SetPitchRange", (PyCFunction)Shake_SetPitchRange, METH_VARARGS, "SetPitchRange(axis: int, octaves: float)\nAxis X=0, Y=1, Z=2, octave range (for one G)"},
    {"SetGain", (PyCFunction)Shake_SetGain, METH_VARARGS, "SetGain(axis: int, gain: float)\nAxis X=0, Y=1, Z=2, mixer gain"},

    {NULL, NULL}
};

static const PyTypeObject c_Shake_type = {
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size        */
    "shake.Shake",                                /* tp_name        */
    sizeof(Shake_PyObject),                     /* tp_basicsize   */
    0,                                        /* tp_itemsize    */
    /* methods */
    (destructor)Shake_PyObject_dealloc,         /* tp_dealloc     */
    0,                                        /* tp_print       */
    0,                                        /* tp_getattr     */
    0,                                        /* tp_setattr     */
    0,                                        /* tp_compare     */
    0,                                        /* tp_repr        */
    0,                                        /* tp_as_number   */
    0,                                        /* tp_as_sequence */
    0,                                        /* tp_as_mapping  */
    0,                                        /* tp_hash        */
    0,                                        /* tp_call        */
    0,                                        /* tp_str         */
    PyObject_GenericGetAttr,                  /* tp_getattro    */
    0,                                        /* tp_setattro    */
    0,                                        /* tp_as_buffer   */
    Py_TPFLAGS_DEFAULT,                       /* tp_flags       */
                                              /* tp_doc         */
    "Shake()\nConstruct Shake object.",
    0,		                              /* tp_traverse */
    0,		                              /* tp_clear */
    0,		                              /* tp_richcompare */
    0,		                              /* tp_weaklistoffset */
    0,		                              /* tp_iter */
    0,		                              /* tp_iternext */
    (PyMethodDef*)Shake_methods,                /* tp_methods */
    0,                                        /* tp_members */
    0,                                        /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    0,                                        /* tp_descr_get */
    0,                                        /* tp_descr_set */
    0,                                        /* tp_dictoffset */
    0,                                        /* tp_init */
    0,                                        /* tp_alloc */
    Shake_PyObject_new,                         /* tp_new */
};

PyTypeObject* Shake_PyType(void)
{
#ifdef __SYMBIAN32__
    static const char* const global_name = "shake_Shake_PyType";
    PyTypeObject* type = (PyTypeObject*)SPyGetGlobalString((char*)global_name);
#else
    static PyTypeObject* type = 0;
#endif
    if (type) {
        return type;
    }
    type = PyObject_New(PyTypeObject, &PyType_Type);
    *type = c_Shake_type;
    type->ob_type = &PyType_Type;
    if (PyType_Ready(type) < 0) {
        type=NULL; // FIXME?
    }
#ifdef __SYMBIAN32__
    SPyAddGlobalString((char*)global_name, (PyObject*)type);
#endif
    return type;
}

static PyObject* Shake_PyObject_new(PyTypeObject* type, PyObject* /*args*/, PyObject* /*kwds*/)
{
    Shake_PyObject* self = (Shake_PyObject*)(type->tp_alloc(type, 0));
    if (!self) {
        return PyErr_NoMemory();
    }
    self->shake=new Shake; // FIXME may throw
    return (PyObject*)self;
}

static void Shake_PyObject_dealloc(Shake_PyObject* self)
{
    delete self->shake;
    PyObject_Del(self);
}

static PyObject* Shake_Open(Shake_PyObject* self, PyObject* args)
{
    if (!PyArg_ParseTuple(args, ":Open")) {
        return NULL;
    }
    int error=0;
    error=self->shake->Open();
    if (error) {
        return SPyErr_SetFromSymbianOSErr(error);
    }
    Py_INCREF(Py_None); return Py_None;
}

static PyObject* Shake_Close(Shake_PyObject* self)
{
    self->shake->Close();
    Py_INCREF(Py_None); return Py_None;
}

static PyObject* Shake_SetVolume(Shake_PyObject* self, PyObject* args)
{
    double vol;
    if (!PyArg_ParseTuple(args, "d:SetVolume", &vol)) {
        return NULL;
    }
    self->shake->SetVolume(vol);
    Py_INCREF(Py_None); return Py_None;
}

static PyObject* Shake_SetFrequency(Shake_PyObject* self, PyObject* args)
{
    int axis;
    double freq;
    if (!PyArg_ParseTuple(args, "id:SetFrequency", &axis, &freq)) {
        return NULL;
    }
    self->shake->SetFrequency(axis, freq);
    Py_INCREF(Py_None); return Py_None;
}

static PyObject* Shake_SetPitchRange(Shake_PyObject* self, PyObject* args)
{
    int axis;
    double range;
    if (!PyArg_ParseTuple(args, "id:SetPitchRange", &axis, &range)) {
        return NULL;
    }
    self->shake->SetPitchRange(axis, range);
    Py_INCREF(Py_None); return Py_None;
}

static PyObject* Shake_SetGain(Shake_PyObject* self, PyObject* args)
{
    int axis;
    double gain;
    if (!PyArg_ParseTuple(args, "id:SetGain", &axis, &gain)) {
        return NULL;
    }
    self->shake->SetGain(axis, gain);
    Py_INCREF(Py_None); return Py_None;
}

DL_EXPORT(void) initshake(void)
{
    PyObject* mod = Py_InitModule3("shake", (PyMethodDef*)shake_methods, "shake C++ audio engine.");
    PyModule_AddObject(mod, "Shake", (PyObject*)Shake_PyType());
}

DL_EXPORT(void) zfinalizeshake(void)
{}

} // extern "C"

#if defined(__SYMBIAN32__) && !defined(EKA2)
GLDEF_C TInt E32Dll(TDllReason aReason)
{
    return KErrNone;
}
#endif

// EOF
